from __future__ import annotations

from collections.abc import Callable, Iterable
from dataclasses import dataclass, field, replace
from functools import partial, wraps
from getpass import getuser
from inspect import iscoroutinefunction, signature
from logging import Formatter, Handler, LogRecord
from pathlib import Path
from socket import gethostname
from sys import exc_info
from textwrap import indent
from traceback import FrameSummary, TracebackException, format_exception
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    Self,
    TypeAlias,
    TypeGuard,
    TypeVar,
    assert_never,
    cast,
    overload,
    runtime_checkable,
)

from typing_extensions import override

from utilities.datetime import get_now
from utilities.errors import ImpossibleCaseError
from utilities.functions import (
    ensure_not_none,
    ensure_str,
    get_class_name,
    get_func_name,
    get_func_qualname,
)
from utilities.git import MASTER
from utilities.iterables import one
from utilities.rich import (
    EXPAND_ALL,
    INDENT_SIZE,
    MAX_DEPTH,
    MAX_LENGTH,
    MAX_STRING,
    MAX_WIDTH,
    yield_call_args_repr,
    yield_mapping_repr,
)
from utilities.version import get_version
from utilities.whenever import serialize_zoned_datetime

if TYPE_CHECKING:
    from collections.abc import Iterator
    from logging import _FormatStyle
    from types import FrameType, TracebackType

    from utilities.types import StrMapping


_F = TypeVar("_F", bound=Callable[..., Any])
_T = TypeVar("_T")
_TExc = TypeVar("_TExc", bound=BaseException)
_CALL_ARGS = "_CALL_ARGS"
_INDENT = 4 * " "


class RichTracebackFormatter(Formatter):
    """Formatter for rich tracebacks."""

    @override
    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: _FormatStyle = "%",
        validate: bool = True,
        /,
        *,
        defaults: StrMapping | None = None,
        git_ref: str = MASTER,
        detail: bool = False,
        post: Callable[[str], str] | None = None,
    ) -> None:
        super().__init__(fmt, datefmt, style, validate, defaults=defaults)
        self._git_ref = git_ref
        self._detail = detail
        self._post = post

    @override
    def format(self, record: LogRecord) -> str:
        """Format the record."""
        if record.exc_info is None:
            return f"ERROR: {record.exc_info=}"
        _, exc_value, _ = record.exc_info
        exc_value = ensure_not_none(exc_value, desc="exc_value")
        error = get_rich_traceback(exc_value, git_ref=self._git_ref)
        match error:
            case ExcChainTB() | ExcGroupTB() | ExcTB():
                text = error.format(header=True, detail=self._detail)
            case BaseException():
                text = "\n".join(format_exception(error))
            case _ as never:
                assert_never(never)
        if self._post is not None:
            text = self._post(text)
        return text

    @classmethod
    def create_and_set(
        cls,
        handler: Handler,
        /,
        *,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: _FormatStyle = "%",
        validate: bool = True,
        defaults: StrMapping | None = None,
        git_ref: str = MASTER,
        detail: bool = False,
        post: Callable[[str], str] | None = None,
    ) -> Self:
        """Create an instance and set it on a handler."""
        formatter = cls(
            fmt,
            datefmt,
            style,
            validate,
            defaults=defaults,
            git_ref=git_ref,
            detail=detail,
            post=post,
        )
        handler.addFilter(lambda r: r.exc_info is not None)
        handler.setFormatter(formatter)
        return formatter


@dataclass(repr=False, kw_only=True, slots=True)
class _CallArgs:
    """A collection of call arguments."""

    func: Callable[..., Any]
    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: dict[str, Any] = field(default_factory=dict)

    @override
    def __repr__(self) -> str:
        cls = get_class_name(self)
        parts: list[tuple[str, Any]] = [
            ("func", get_func_qualname(self.func)),
            ("args", self.args),
            ("kwargs", self.kwargs),
        ]
        joined = ", ".join(f"{k}={v!r}" for k, v in parts)
        return f"{cls}({joined})"

    @classmethod
    def create(cls, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Self:
        """Make the initial trace data."""
        sig = signature(func)
        try:
            bound_args = sig.bind(*args, **kwargs)
        except TypeError as error:
            orig = ensure_str(one(error.args))
            lines: list[str] = [
                f"Unable to bind arguments for {get_func_name(func)!r}; {orig}"
            ]
            lines.extend(yield_call_args_repr(*args, **kwargs))
            new = "\n".join(lines)
            raise _CallArgsError(new) from None
        return cls(func=func, args=bound_args.args, kwargs=bound_args.kwargs)


class _CallArgsError(TypeError):
    """Raised when a set of call arguments cannot be created."""


@dataclass(kw_only=True, slots=True)
class _ExtFrameSummary(Generic[_T]):
    """An extended frame summary."""

    filename: Path
    module: str | None = None
    name: str
    qualname: str
    code_line: str
    first_line_num: int
    line_num: int
    end_line_num: int
    col_num: int | None = None
    end_col_num: int | None = None
    locals: dict[str, Any] = field(default_factory=dict)
    extra: _T


_ExtFrameSummaryCAOpt: TypeAlias = _ExtFrameSummary[_CallArgs | None]
_ExtFrameSummaryCA: TypeAlias = _ExtFrameSummary[_CallArgs]


@dataclass(repr=False, kw_only=True, slots=True)
class _ExcTBInternal:
    """A rich traceback for an exception; internal use only."""

    raw: list[_ExtFrameSummaryCAOpt] = field(default_factory=list)
    frames: list[_ExtFrameSummaryCA] = field(default_factory=list)
    error: BaseException


@runtime_checkable
class _HasExceptionPath(Protocol):
    @property
    def exc_tb(self) -> _ExcTBInternal: ...  # pragma: no cover


@dataclass(kw_only=True, slots=True)
class ExcChainTB(Generic[_TExc]):
    """A rich traceback for an exception chain."""

    errors: list[ExcGroupTB[_TExc] | ExcTB[_TExc] | BaseException] = field(
        default_factory=list
    )
    git_ref: str = field(default=MASTER, repr=False)

    def __iter__(self) -> Iterator[ExcGroupTB[_TExc] | ExcTB[_TExc] | BaseException]:
        yield from self.errors

    def __len__(self) -> int:
        return len(self.errors)

    @override
    def __repr__(self) -> str:
        return self.format(header=True, detail=True)

    def format(
        self,
        *,
        header: bool = False,
        detail: bool = False,
        max_width: int = MAX_WIDTH,
        indent_size: int = INDENT_SIZE,
        max_length: int | None = MAX_LENGTH,
        max_string: int | None = MAX_STRING,
        max_depth: int | None = MAX_DEPTH,
        expand_all: bool = EXPAND_ALL,
    ) -> str:
        """Format the traceback."""
        lines: list[str] = []
        if header:  # pragma: no cover
            lines.extend(_yield_header_lines(git_ref=self.git_ref))
        total = len(self.errors)
        for i, errors in enumerate(self.errors):
            lines.append(f"Exception chain {i + 1}/{total}:")
            match errors:
                case ExcGroupTB():  # pragma: no cover
                    lines.append(
                        errors.format(
                            index=i,
                            total=total,
                            header=False,
                            detail=detail,
                            max_width=max_width,
                            indent_size=indent_size,
                            max_length=max_length,
                            max_string=max_string,
                            max_depth=max_depth,
                            expand_all=expand_all,
                            depth=1,
                        )
                    )
                case ExcTB():  # pragma: no cover
                    lines.append(
                        errors.format(
                            header=False,
                            detail=detail,
                            max_width=max_width,
                            indent_size=indent_size,
                            max_length=max_length,
                            max_string=max_string,
                            max_depth=max_depth,
                            expand_all=expand_all,
                            depth=1,
                        )
                    )
                case BaseException():  # pragma: no cover
                    lines.append(_format_exception(errors, depth=1))
                case _ as never:
                    assert_never(never)
            lines.append("")
        return "\n".join(lines)


@dataclass(kw_only=True, slots=True)
class ExcGroupTB(Generic[_TExc]):
    """A rich traceback for an exception group."""

    path: ExcTB[_TExc]
    errors: list[ExcGroupTB[_TExc] | ExcTB[_TExc] | BaseException] = field(
        default_factory=list
    )
    git_ref: str = field(default=MASTER, repr=False)

    @override
    def __repr__(self) -> str:
        return self.format(header=True, detail=True)

    def format(
        self,
        *,
        index: int = 0,
        total: int = 1,
        header: bool = False,
        detail: bool = False,
        max_width: int = MAX_WIDTH,
        indent_size: int = INDENT_SIZE,
        max_length: int | None = MAX_LENGTH,
        max_string: int | None = MAX_STRING,
        max_depth: int | None = MAX_DEPTH,
        expand_all: bool = EXPAND_ALL,
        depth: int = 0,
    ) -> str:
        """Format the traceback."""
        lines: list[str] = []
        if header:  # pragma: no cover
            lines.extend(_yield_header_lines(git_ref=self.git_ref))
        lines.extend([
            f"Exception group {index + 1}/{total}:",
            indent("Path:", _INDENT),
            self.path.format(header=False, detail=detail, depth=2),
            "",
        ])
        total_sub_errors = len(self.errors)
        for i, errors in enumerate(self.errors):
            lines.append(indent(f"Group error {i + 1}/{total_sub_errors}:", _INDENT))
            match errors:
                case ExcGroupTB() | ExcTB():  # pragma: no cover
                    lines.append(
                        errors.format(
                            header=False,
                            detail=detail,
                            max_width=max_width,
                            indent_size=indent_size,
                            max_length=max_length,
                            max_string=max_string,
                            max_depth=max_depth,
                            expand_all=expand_all,
                            depth=2,
                        )
                    )
                case BaseException():  # pragma: no cover
                    lines.append(_format_exception(errors, depth=2))
                case _ as never:
                    assert_never(never)
            lines.append("")
        return indent("\n".join(lines), depth * _INDENT)


@dataclass(kw_only=True, slots=True)
class ExcTB(Generic[_TExc]):
    """A rich traceback for a single exception."""

    frames: list[_Frame] = field(default_factory=list)
    error: _TExc
    git_ref: str = field(default=MASTER, repr=False)

    def __iter__(self) -> Iterator[_Frame]:
        yield from self.frames

    def __len__(self) -> int:
        return len(self.frames)

    @override
    def __repr__(self) -> str:
        return self.format(header=True, detail=True)

    def format(
        self,
        *,
        header: bool = False,
        detail: bool = False,
        max_width: int = MAX_WIDTH,
        indent_size: int = INDENT_SIZE,
        max_length: int | None = MAX_LENGTH,
        max_string: int | None = MAX_STRING,
        max_depth: int | None = MAX_DEPTH,
        expand_all: bool = EXPAND_ALL,
        depth: int = 0,
    ) -> str:
        """Format the traceback."""
        total = len(self)
        lines: list[str] = []
        if header:  # pragma: no cover
            lines.extend(_yield_header_lines(git_ref=self.git_ref))
        for i, frame in enumerate(self.frames):
            is_head = i < total - 1
            lines.append(
                frame.format(
                    index=i,
                    total=total,
                    detail=detail,
                    error=None if is_head else self.error,
                    max_width=max_width,
                    indent_size=indent_size,
                    max_length=max_length,
                    max_string=max_string,
                    max_depth=max_depth,
                    expand_all=expand_all,
                )
            )
            if detail and is_head:
                lines.append("")
        return indent("\n".join(lines), depth * _INDENT)


@dataclass(kw_only=True, slots=True)
class _Frame:
    module: str | None = None
    name: str
    code_line: str
    line_num: int
    args: tuple[Any, ...] = field(default_factory=tuple)
    kwargs: dict[str, Any] = field(default_factory=dict)
    locals: dict[str, Any] = field(default_factory=dict)

    @override
    def __repr__(self) -> str:
        return self.format(detail=True)

    def format(
        self,
        *,
        index: int = 0,
        total: int = 1,
        detail: bool = False,
        error: BaseException | None = None,
        depth: int = 0,
        max_width: int = MAX_WIDTH,
        indent_size: int = INDENT_SIZE,
        max_length: int | None = MAX_LENGTH,
        max_string: int | None = MAX_STRING,
        max_depth: int | None = MAX_DEPTH,
        expand_all: bool = EXPAND_ALL,
    ) -> str:
        """Format the traceback."""
        lines: list[str] = [f"Frame {index + 1}/{total}: {self.name} ({self.module})"]
        if detail:
            lines.append(indent("Inputs:", _INDENT))
            lines.extend(
                indent(line, 2 * _INDENT)
                for line in yield_call_args_repr(
                    *self.args,
                    _max_width=max_width,
                    _indent_size=indent_size,
                    _max_length=max_length,
                    _max_string=max_string,
                    _max_depth=max_depth,
                    _expand_all=expand_all,
                    **self.kwargs,
                )
            )
            lines.append(indent("Locals:", _INDENT))
            lines.extend(
                indent(line, 2 * _INDENT)
                for line in yield_mapping_repr(
                    _max_width=max_width,
                    _indent_size=indent_size,
                    _max_length=max_length,
                    _max_string=max_string,
                    _max_depth=max_depth,
                    _expand_all=expand_all,
                    **self.locals,
                )
            )
            lines.extend([
                indent(f"Line {self.line_num}:", _INDENT),
                indent(self.code_line, 2 * _INDENT),
            ])
        if error is not None:
            lines.extend([
                indent("Raised:", _INDENT),
                _format_exception(error, depth=2),
            ])
        return indent("\n".join(lines), depth * _INDENT)


def get_rich_traceback(
    error: _TExc, /, *, git_ref: str = MASTER
) -> ExcChainTB[_TExc] | ExcGroupTB[_TExc] | ExcTB[_TExc] | BaseException:
    """Get a rich traceback."""
    match list(yield_exceptions(error)):
        case []:  # pragma: no cover
            raise ImpossibleCaseError(case=[f"{error}"])
        case [err]:
            err = cast(_TExc, err)
            return _get_rich_traceback_non_chain(err, git_ref=git_ref)
        case errors:
            errors = cast(list[_TExc], errors)
            return ExcChainTB(
                errors=[
                    _get_rich_traceback_non_chain(e, git_ref=git_ref) for e in errors
                ],
                git_ref=git_ref,
            )


def _get_rich_traceback_non_chain(
    error: _TExc, /, *, git_ref: str = MASTER
) -> ExcTB[_TExc] | ExcGroupTB[_TExc] | BaseException:
    """Get a rich traceback, for a non-chained error."""
    if not isinstance(error, ExceptionGroup):
        return _get_rich_traceback_non_chain_non_group(error, git_ref=git_ref)
    path = cast(
        ExcTB[_TExc], _get_rich_traceback_non_chain_non_group(error, git_ref=git_ref)
    )
    errors = cast(list[_TExc], error.exceptions)
    errors = [_get_rich_traceback_non_chain(e, git_ref=git_ref) for e in errors]
    return ExcGroupTB(path=path, errors=errors, git_ref=git_ref)


def _get_rich_traceback_non_chain_non_group(
    error: _TExc, /, *, git_ref: str = MASTER
) -> ExcTB[_TExc] | BaseException:
    """Get a rich traceback, for a non-chained, non-grouped error."""
    if isinstance(error, _HasExceptionPath):
        frames = [
            _Frame(
                module=f.module,
                name=f.name,
                code_line=f.code_line,
                line_num=f.line_num,
                args=f.extra.args,
                kwargs=f.extra.kwargs,
                locals=f.locals,
            )
            for f in error.exc_tb.frames
        ]
        return ExcTB(frames=frames, error=cast(_TExc, error), git_ref=git_ref)
    return error


@overload
def trace(
    func: _F,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> _F: ...
@overload
def trace(
    func: None = None,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> Callable[[_F], _F]: ...
def trace(
    func: _F | None = None,
    /,
    *,
    setup: Callable[[], bool] | None = None,
    runtime: Callable[[], bool] | None = None,
) -> _F | Callable[[_F], _F]:
    """Trace a function call."""
    if func is None:
        result = partial(trace, setup=setup, runtime=runtime)
        return cast(Callable[[_F], _F], result)

    if (setup is not None) and not setup():
        return func

    if runtime is None:
        if not iscoroutinefunction(func):

            @wraps(func)
            def trace_sync(*args: Any, **kwargs: Any) -> Any:
                locals()[_CALL_ARGS] = _CallArgs.create(func, *args, **kwargs)
                try:
                    return func(*args, **kwargs)
                except Exception as error:
                    cast(Any, error).exc_tb = _get_rich_traceback_internal(error)
                    raise

            return cast(_F, trace_sync)

        @wraps(func)
        async def trace_async(*args: Any, **kwargs: Any) -> Any:
            locals()[_CALL_ARGS] = _CallArgs.create(func, *args, **kwargs)
            try:
                return await func(*args, **kwargs)
            except Exception as error:
                cast(Any, error).exc_tb = _get_rich_traceback_internal(error)
                raise

        return cast(_F, trace_async)

    if not iscoroutinefunction(func):

        @wraps(func)
        def trace_sync(*args: Any, **kwargs: Any) -> Any:
            if en := runtime():
                locals()[_CALL_ARGS] = _CallArgs.create(func, *args, **kwargs)
            try:
                return func(*args, **kwargs)
            except Exception as error:
                if en:
                    cast(Any, error).exc_tb = _get_rich_traceback_internal(error)
                raise

        return cast(_F, trace_sync)

    @wraps(func)
    async def trace_async(*args: Any, **kwargs: Any) -> Any:
        if en := runtime():
            locals()[_CALL_ARGS] = _CallArgs.create(func, *args, **kwargs)
        try:
            return await func(*args, **kwargs)
        except Exception as error:
            if en:
                cast(Any, error).exc_tb = _get_rich_traceback_internal(error)
            raise

    return cast(_F, trace_async)


@overload
def yield_extended_frame_summaries(
    error: BaseException, /, *, extra: Callable[[FrameSummary, FrameType], _T]
) -> Iterator[_ExtFrameSummary[_T]]: ...
@overload
def yield_extended_frame_summaries(
    error: BaseException, /, *, extra: None = None
) -> Iterator[_ExtFrameSummary[None]]: ...
def yield_extended_frame_summaries(
    error: BaseException,
    /,
    *,
    extra: Callable[[FrameSummary, FrameType], _T] | None = None,
) -> Iterator[_ExtFrameSummary[Any]]:
    """Yield the extended frame summaries."""
    tb_exc = TracebackException.from_exception(error, capture_locals=True)
    _, _, traceback = exc_info()
    frames = yield_frames(traceback=traceback)
    for summary, frame in zip(tb_exc.stack, frames, strict=True):
        if extra is None:
            extra_use: _T | None = None
        else:
            extra_use: _T | None = extra(summary, frame)
        yield _ExtFrameSummary(
            filename=Path(summary.filename),
            module=frame.f_globals.get("__name__"),
            name=summary.name,
            qualname=frame.f_code.co_qualname,
            code_line=ensure_not_none(summary.line, desc="summary.line"),
            first_line_num=frame.f_code.co_firstlineno,
            line_num=ensure_not_none(summary.lineno, desc="summary.lineno"),
            end_line_num=ensure_not_none(summary.end_lineno, desc="summary.end_lineno"),
            col_num=summary.colno,
            end_col_num=summary.end_colno,
            locals=frame.f_locals,
            extra=extra_use,
        )


def yield_exceptions(error: BaseException, /) -> Iterator[BaseException]:
    """Yield the exceptions in a context chain."""
    curr: BaseException | None = error
    while curr is not None:
        yield curr
        curr = curr.__context__


def yield_frames(*, traceback: TracebackType | None = None) -> Iterator[FrameType]:
    """Yield the frames of a traceback."""
    while traceback is not None:
        yield traceback.tb_frame
        traceback = traceback.tb_next


def _format_exception(error: BaseException, /, *, depth: int = 0) -> str:
    """Format an exception."""
    cls = get_class_name(error)
    line = f"{cls}({error})"
    return indent(line, depth * _INDENT)


def _get_rich_traceback_internal(error: BaseException, /) -> _ExcTBInternal:
    """Get a rich traceback; for internal use only."""

    def extra(_: FrameSummary, frame: FrameType) -> _CallArgs | None:
        return frame.f_locals.get(_CALL_ARGS)

    raw = list(yield_extended_frame_summaries(error, extra=extra))
    return _ExcTBInternal(raw=raw, frames=_merge_frames(raw), error=error)


def _merge_frames(
    frames: Iterable[_ExtFrameSummaryCAOpt], /
) -> list[_ExtFrameSummaryCA]:
    """Merge a set of frames."""
    rev = list(frames)[::-1]
    values: list[_ExtFrameSummaryCA] = []

    def get_solution(
        curr: _ExtFrameSummaryCAOpt, rev: list[_ExtFrameSummaryCAOpt], /
    ) -> _ExtFrameSummaryCA:
        while True:
            next_ = rev.pop(0)
            if has_extra(next_) and is_match(curr, next_):
                return next_

    def has_extra(frame: _ExtFrameSummaryCAOpt, /) -> TypeGuard[_ExtFrameSummaryCA]:
        return frame.extra is not None

    def has_match(
        curr: _ExtFrameSummaryCAOpt, rev: list[_ExtFrameSummaryCAOpt], /
    ) -> bool:
        next_, *_ = filter(has_extra, rev)
        return is_match(curr, next_)

    def is_match(curr: _ExtFrameSummaryCAOpt, next_: _ExtFrameSummaryCA, /) -> bool:
        return (curr.name == next_.extra.func.__name__) and (
            (curr.module is None) or (curr.module == next_.extra.func.__module__)
        )

    while len(rev) >= 1:
        curr = rev.pop(0)
        if not has_match(curr, rev):
            continue
        next_ = get_solution(curr, rev)
        new = cast(_ExtFrameSummaryCA, replace(curr, extra=next_.extra))
        values.append(new)
    return values[::-1]


def _yield_header_lines(*, git_ref: str = MASTER) -> Iterator[str]:
    """Yield the header lines."""
    yield f"Date/time | {serialize_zoned_datetime(get_now(time_zone='local'))}"
    yield f"User      | {getuser()}"
    yield f"Host      | {gethostname()}"
    yield f"Version   | {get_version(ref=git_ref)}"
    yield ""


__all__ = [
    "ExcChainTB",
    "ExcGroupTB",
    "ExcTB",
    "RichTracebackFormatter",
    "get_rich_traceback",
    "trace",
    "yield_exceptions",
    "yield_extended_frame_summaries",
    "yield_frames",
]



from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from typing_extensions import override

if TYPE_CHECKING:
    from utilities.types import ZoneInfoLike

HongKong = ZoneInfo("Asia/Hong_Kong")
Tokyo = ZoneInfo("Asia/Tokyo")
USCentral = ZoneInfo("US/Central")
USEastern = ZoneInfo("US/Eastern")
UTC = ZoneInfo("UTC")


def ensure_time_zone(time_zone: ZoneInfoLike | dt.tzinfo, /) -> ZoneInfo:
    """Ensure the object is a time zone."""
    if isinstance(time_zone, ZoneInfo):
        return time_zone
    if isinstance(time_zone, str):
        return ZoneInfo(time_zone)
    if time_zone is dt.UTC:
        return UTC
    raise EnsureTimeZoneError(time_zone=time_zone)


@dataclass(kw_only=True, slots=True)
class EnsureTimeZoneError(Exception):
    time_zone: dt.tzinfo

    @override
    def __str__(self) -> str:
        return f"Unsupported time zone: {self.time_zone}"


def get_time_zone_name(time_zone: ZoneInfoLike | dt.timezone, /) -> str:
    """Get the name of a time zone."""
    return ensure_time_zone(time_zone).key


__all__ = [
    "UTC",
    "EnsureTimeZoneError",
    "HongKong",
    "Tokyo",
    "USCentral",
    "USEastern",
    "ensure_time_zone",
    "get_time_zone_name",
]



from __future__ import annotations

import re
from os import environ
from re import MULTILINE, Pattern

from pytest import fixture, mark

from utilities.platform import IS_NOT_LINUX, IS_WINDOWS
from utilities.text import strip_and_dedent

FLAKY = mark.flaky(reruns=5, reruns_delay=1)
IS_CI = "CI" in environ
SKIPIF_CI = mark.skipif(IS_CI, reason="Skipped for CI")
IS_CI_AND_WINDOWS = IS_CI and IS_WINDOWS
SKIPIF_CI_AND_WINDOWS = mark.skipif(IS_CI_AND_WINDOWS, reason="Skipped for CI/Windows")
SKIPIF_CI_AND_NOT_LINUX = mark.skipif(
    IS_CI and IS_NOT_LINUX, reason="Skipped for CI/non-Linux"
)


# hypothesis


try:
    from utilities.hypothesis import setup_hypothesis_profiles
except ModuleNotFoundError:
    pass
else:
    setup_hypothesis_profiles()


# fixtures


@fixture
def git_ref() -> str:
    return "origin/master" if IS_CI else "master"


@fixture
def traceback_func_chain() -> Pattern[str]:
    return re.compile(
        strip_and_dedent(
            r"""
            Date/time \| .+
            User      \| .+
            Host      \| .+
            Version   \| .+

            Exception chain 1/2:
                Frame 1/1: func_chain_first \(tests\.test_traceback_funcs\.chain\)
                    Inputs:
                        args\[0\] = 1
                        args\[1\] = 2
                        args\[2\] = 3
                        args\[3\] = 4
                        kwargs\[c\] = 5
                        kwargs\[d\] = 6
                        kwargs\[e\] = 7
                    Locals:
                        a = 2
                        b = 4
                        c = 10
                        args = \(6, 8\)
                        kwargs = {'d': 12, 'e': 14}
                        msg = 'Assertion failed: Result \(112\) must be divisible by 10'
                    Line 19:
                        raise ValueError\(msg\) from error
                    Raised:
                        ValueError\(Assertion failed: Result \(112\) must be divisible by 10\)

            Exception chain 2/2:
                Frame 1/1: func_chain_second \(tests\.test_traceback_funcs\.chain\)
                    Inputs:
                        args\[0\] = 2
                        args\[1\] = 4
                        args\[2\] = 6
                        args\[3\] = 8
                        kwargs\[c\] = 10
                        kwargs\[d\] = 12
                        kwargs\[e\] = 14
                    Locals:
                        a = 4
                        b = 8
                        c = 20
                        args = \(12, 16\)
                        kwargs = {'d': 24, 'e': 28}
                        result = 112
                    Line 30:
                        assert result % 10 == 0, f"Result \({result}\) must be divisible by 10"
                    Raised:
                        AssertionError\(Result \(112\) must be divisible by 10\)
            """
        ),
        flags=MULTILINE,
    )


@fixture
def traceback_func_one() -> Pattern[str]:
    return re.compile(
        strip_and_dedent(
            r"""
            Date/time \| .+
            User      \| .+
            Host      \| .+
            Version   \| .+

            Frame 1/1: func_one \(tests\.test_traceback_funcs\.one\)
                Inputs:
                    args\[0\] = 1
                    args\[1\] = 2
                    args\[2\] = 3
                    args\[3\] = 4
                    kwargs\[c\] = 5
                    kwargs\[d\] = 6
                    kwargs\[e\] = 7
                Locals:
                    a = 2
                    b = 4
                    c = 10
                    args = \(6, 8\)
                    kwargs = {'d': 12, 'e': 14}
                    result = 56
                Line 16:
                    assert result % 10 == 0, f"Result \({result}\) must be divisible by 10"
                Raised:
                    AssertionError\(Result \(56\) must be divisible by 10\)
            """
        ),
        flags=MULTILINE,
    )


@fixture
def traceback_func_task_group_one() -> Pattern[str]:
    return re.compile(
        strip_and_dedent(
            r"""
            Date/time \| .+
            User      \| .+
            Host      \| .+
            Version   \| .+

            Exception group 1/1:
                Path:
                    Frame 1/1: func_task_group_one_first \(tests\.test_traceback_funcs\.task_group_one\)
                        Inputs:
                            args\[0\] = 1
                            args\[1\] = 2
                            args\[2\] = 3
                            args\[3\] = 4
                            kwargs\[c\] = 5
                            kwargs\[d\] = 6
                            kwargs\[e\] = 7
                        Locals:
                            a = 2
                            b = 4
                            c = 10
                            args = \(6, 8\)
                            kwargs = {'d': 12, 'e': 14}
                            tg = <TaskGroup cancelling>
                            _ = <Task finished name='Task-\d+' coro=<func_task_group_one_second\(\) done, defined at .+src.+utilities.+traceback\.py:\d+> exception=AssertionError\('Result \(112\) must be divisible by 10'\)>
                        Line 18:
                            async with TaskGroup\(\) as tg:
                        Raised:
                            ExceptionGroup\(unhandled errors in a TaskGroup \(1 sub-exception\)\)

                Group error 1/1:
                    Frame 1/1: func_task_group_one_second \(tests\.test_traceback_funcs\.task_group_one\)
                        Inputs:
                            args\[0\] = 2
                            args\[1\] = 4
                            args\[2\] = 6
                            args\[3\] = 8
                            kwargs\[c\] = 10
                            kwargs\[d\] = 12
                            kwargs\[e\] = 14
                        Locals:
                            a = 4
                            b = 8
                            c = 20
                            args = \(12, 16\)
                            kwargs = {'d': 24, 'e': 28}
                            result = 112
                        Line 33:
                            assert result % 10 == 0, f"Result \({result}\) must be divisible by 10"
                        Raised:
                            AssertionError\(Result \(112\) must be divisible by 10\)
            """
        )
    )


@fixture
def traceback_func_two() -> Pattern[str]:
    return re.compile(
        strip_and_dedent(
            r"""
            Date/time \| .+
            User      \| .+
            Host      \| .+
            Version   \| .+

            Frame 1/2: func_two_first \(tests\.test_traceback_funcs\.two\)
                Inputs:
                    args\[0\] = 1
                    args\[1\] = 2
                    args\[2\] = 3
                    args\[3\] = 4
                    kwargs\[c\] = 5
                    kwargs\[d\] = 6
                    kwargs\[e\] = 7
                Locals:
                    a = 2
                    b = 4
                    c = 10
                    args = \(6, 8\)
                    kwargs = {'d': 12, 'e': 14}
                Line 15:
                    return func_two_second\(a, b, \*args, c=c, \*\*kwargs\)

            Frame 2/2: func_two_second \(tests\.test_traceback_funcs\.two\)
                Inputs:
                    args\[0\] = 2
                    args\[1\] = 4
                    args\[2\] = 6
                    args\[3\] = 8
                    kwargs\[c\] = 10
                    kwargs\[d\] = 12
                    kwargs\[e\] = 14
                Locals:
                    a = 4
                    b = 8
                    c = 20
                    args = \(12, 16\)
                    kwargs = {'d': 24, 'e': 28}
                    result = 112
                Line 26:
                    assert result % 10 == 0, f"Result \({result}\) must be divisible by 10"
                Raised:
                    AssertionError\(Result \(112\) must be divisible by 10\)
            """
        ),
        flags=MULTILINE,
    )


@fixture
def traceback_func_untraced() -> Pattern[str]:
    return re.compile(
        strip_and_dedent(
            r"""
            Traceback \(most recent call last\):

              File ".+src.+tests.+test_(logging|sys|traceback)\.py", line \d+, in test_.+
                _ = func_untraced\(1, 2, 3, 4, c=5, d=6, e=7\)
                    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

              File ".+src.+tests.+test_traceback_funcs.+untraced\.py", line 13, in func_untraced
                assert result % 10 == 0, f"Result \({result}\) must be divisible by 10"
                       ^^^^^^^^^^^^^^^^

            AssertionError: Result \(56\) must be divisible by 10
            """
        ).replace("^", r"\^"),
        flags=MULTILINE,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from utilities.types import Dataclass

if TYPE_CHECKING:
    from collections.abc import Callable

    from optuna import Study, Trial


_TDataclass = TypeVar("_TDataclass", bound=Dataclass)


def get_best_params(study: Study, cls: type[_TDataclass], /) -> _TDataclass:
    """Get the best params as a dataclass."""
    return cls(**study.best_params)


def make_objective(
    suggest_params: Callable[[Trial], _TDataclass],
    objective: Callable[[_TDataclass], float],
    /,
) -> Callable[[Trial], float]:
    """Make an objective given separate trialling & evaluating functions."""

    def inner(trial: Trial, /) -> float:
        return objective(suggest_params(trial))

    return inner


def suggest_bool(trial: Trial, name: str, /) -> bool:
    """Suggest a boolean value."""
    return trial.suggest_categorical(name, [True, False])


__all__ = ["get_best_params", "make_objective", "suggest_bool"]



from __future__ import annotations

import datetime as dt
from collections.abc import Awaitable, Callable, Coroutine, Hashable, Iterable, Mapping
from enum import Enum
from logging import Logger
from pathlib import Path
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    Literal,
    Protocol,
    TypeAlias,
    TypeVar,
    runtime_checkable,
)
from zoneinfo import ZoneInfo

_T = TypeVar("_T")
_TEnum = TypeVar("_TEnum", bound=Enum)
_THashable = TypeVar("_THashable", bound=Hashable)
_T_contra = TypeVar("_T_contra", contravariant=True)


# basic
Number: TypeAlias = int | float
Duration: TypeAlias = Number | dt.timedelta
StrMapping: TypeAlias = Mapping[str, Any]
TupleOrStrMapping: TypeAlias = tuple[Any, ...] | StrMapping


# asyncio
Coroutine1: TypeAlias = Coroutine[Any, Any, _T]
MaybeAwaitable: TypeAlias = _T | Awaitable[_T]
MaybeCoroutine1: TypeAlias = _T | Coroutine1[_T]


# dataclasses
@runtime_checkable
class Dataclass(Protocol):
    """Protocol for `dataclass` classes."""

    __dataclass_fields__: ClassVar[dict[str, Any]]


# enum
EnumOrStr: TypeAlias = _TEnum | str


# iterables
MaybeIterable: TypeAlias = _T | Iterable[_T]
IterableHashable: TypeAlias = tuple[_THashable, ...] | frozenset[_THashable]
MaybeIterableHashable: TypeAlias = _THashable | IterableHashable[_THashable]


# logging
LogLevel: TypeAlias = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LoggerOrName: TypeAlias = Logger | str


# operator


class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra, /) -> bool: ...  # pragma: no cover


class SupportsDunderGT(Protocol[_T_contra]):
    def __gt__(self, other: _T_contra, /) -> bool: ...  # pragma: no cover


SupportsRichComparison = SupportsDunderLT[Any] | SupportsDunderGT[Any]


# pathlib
PathLike: TypeAlias = Path | str
PathLikeOrCallable: TypeAlias = PathLike | Callable[[], PathLike]


# traceback
ExcInfo: TypeAlias = tuple[type[BaseException], BaseException, TracebackType]
OptExcInfo: TypeAlias = ExcInfo | tuple[None, None, None]


# zoneinfo
ZoneInfoLike: TypeAlias = ZoneInfo | str


__all__ = [
    "Coroutine1",
    "Dataclass",
    "Duration",
    "EnumOrStr",
    "ExcInfo",
    "IterableHashable",
    "LogLevel",
    "LoggerOrName",
    "MaybeAwaitable",
    "MaybeCoroutine1",
    "MaybeIterable",
    "MaybeIterableHashable",
    "Number",
    "OptExcInfo",
    "PathLike",
    "PathLikeOrCallable",
    "StrMapping",
    "SupportsDunderGT",
    "SupportsDunderLT",
    "SupportsRichComparison",
    "TupleOrStrMapping",
    "ZoneInfoLike",
]



from __future__ import annotations

from asyncio import sleep
from re import search
from typing import TYPE_CHECKING, ClassVar, Literal

from eventkit import Event
from hypothesis import HealthCheck, given
from hypothesis.strategies import integers, sampled_from
from pytest import CaptureFixture

from tests.conftest import FLAKY
from utilities.eventkit import add_listener
from utilities.functions import identity
from utilities.hypothesis import settings_with_reduced_examples

if TYPE_CHECKING:
    from pytest import CaptureFixture


class TestAddListener:
    datetime: ClassVar[str] = r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3} \| "

    @FLAKY
    @given(sync_or_async=sampled_from(["sync", "async"]), n=integers())
    @settings_with_reduced_examples(
        suppress_health_check={HealthCheck.function_scoped_fixture}
    )
    async def test_main(
        self, *, sync_or_async: Literal["sync", "async"], n: int
    ) -> None:
        event = Event()
        match sync_or_async:
            case "sync":

                def listener_sync(n: int, /) -> None:
                    print(f"n={n}")  # noqa: T201

                _ = add_listener(event, listener_sync)
            case "async":

                async def listener_async(n: int, /) -> None:
                    await sleep(0.01)
                    print(f"n={n}")  # noqa: T201

                _ = add_listener(event, listener_async)

        event.emit(n)

    @given(n=integers())
    @settings_with_reduced_examples(
        suppress_health_check={HealthCheck.function_scoped_fixture}
    )
    def test_custom_error_handler(self, *, capsys: CaptureFixture, n: int) -> None:
        event = Event()

        def error(event: Event, exception: Exception, /) -> None:
            _ = (event, exception)
            print("Custom handler")  # noqa: T201

        _ = add_listener(event, identity, error=error)
        event.emit(n, n)
        out = capsys.readouterr().out
        *_, line = out.splitlines()
        assert line == "Custom handler"

    @given(n=integers())
    @settings_with_reduced_examples(
        suppress_health_check={HealthCheck.function_scoped_fixture}
    )
    async def test_error_stdout(self, *, capsys: CaptureFixture, n: int) -> None:
        event = Event()
        _ = add_listener(event, identity)
        event.emit(n, n)
        out = capsys.readouterr().out
        (line1, line2, line3) = out.splitlines()
        assert line1 == "Raised a TypeError whilst running 'Event':"
        pattern2 = (
            r"^event=Event<Event, \[\[None, None, <function identity at .*>\]\]>$"
        )
        assert search(pattern2, line2)
        assert (
            line3
            == "exception=TypeError('identity() takes 1 positional argument but 2 were given')"
        )

    @given(n=integers())
    @settings_with_reduced_examples(
        suppress_health_check={HealthCheck.function_scoped_fixture}
    )
    async def test_error_ignore(self, *, capsys: CaptureFixture, n: int) -> None:
        event = Event()
        _ = add_listener(event, identity, error_ignore=TypeError)
        event.emit(n, n)
        out = capsys.readouterr().out
        expected = ""
        assert out == expected



from __future__ import annotations

import re
from dataclasses import dataclass, field
from pathlib import Path
from re import DOTALL
from types import NoneType
from typing import Any, Literal, cast

from hypothesis import given
from hypothesis.strategies import integers, lists, sampled_from
from ib_async import Future
from polars import DataFrame
from pytest import raises
from typing_extensions import override

from tests.test_operator import DataClass3
from utilities.dataclasses import (
    YieldFieldsError,
    _MappingToDataclassCaseInsensitiveNonUniqueError,
    _MappingToDataclassEmptyError,
    _YieldFieldsClass,
    _YieldFieldsInstance,
    asdict_without_defaults,
    mapping_to_dataclass,
    replace_non_sentinel,
    repr_without_defaults,
    yield_fields,
)
from utilities.functions import get_class_name
from utilities.hypothesis import text_ascii
from utilities.iterables import one
from utilities.orjson import OrjsonLogRecord
from utilities.polars import are_frames_equal
from utilities.sentinel import sentinel
from utilities.types import Dataclass, StrMapping
from utilities.typing import get_args, is_list_type, is_literal_type, is_optional_type

TruthLit = Literal["true", "false"]  # in 3.12, use type TruthLit = ...


class TestAsDictWithoutDefaultsAndReprWithoutDefaults:
    @given(x=integers())
    def test_field_without_defaults(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        obj = Example(x=x)
        asdict_res = asdict_without_defaults(obj)
        asdict_exp = {"x": x}
        assert asdict_res == asdict_exp
        repr_res = repr_without_defaults(obj)
        repr_exp = f"Example(x={x})"
        assert repr_res == repr_exp

    def test_field_with_default(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example()
        asdict_res = asdict_without_defaults(obj)
        asdict_exp = {}
        assert asdict_res == asdict_exp
        repr_res = repr_without_defaults(obj)
        repr_exp = "Example()"
        assert repr_res == repr_exp

    def test_field_with_dataframe(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: DataFrame = field(default_factory=DataFrame)

        obj = Example()
        extra = {DataFrame: are_frames_equal}
        asdict_res = asdict_without_defaults(obj, globalns=globals(), extra=extra)
        asdict_exp = {}
        assert set(asdict_res) == set(asdict_exp)
        repr_res = repr_without_defaults(obj, globalns=globals(), extra=extra)
        repr_exp = "Example()"
        assert repr_res == repr_exp

    @given(x=integers())
    def test_final(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        def final(obj: type[Dataclass], mapping: StrMapping) -> StrMapping:
            return {f"[{get_class_name(obj)}]": mapping}

        obj = Example(x=x)
        result = asdict_without_defaults(obj, final=final)
        expected = {"[Example]": {"x": x}}
        assert result == expected

    @given(x=integers())
    def test_nested_with_recursive(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner

        obj = Outer(inner=Inner(x=x))
        asdict_res = asdict_without_defaults(obj, localns=locals(), recursive=True)
        asdict_exp = {"inner": {"x": x}}
        assert asdict_res == asdict_exp
        repr_res = repr_without_defaults(obj, localns=locals(), recursive=True)
        repr_exp = f"Outer(inner=Inner(x={x}))"
        assert repr_res == repr_exp

    @given(x=integers())
    def test_nested_without_recursive(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner

        obj = Outer(inner=Inner(x=x))
        asdict_res = asdict_without_defaults(obj, localns=locals())
        asdict_exp = {"inner": Inner(x=x)}
        assert asdict_res == asdict_exp
        repr_res = repr_without_defaults(obj, localns=locals())
        repr_exp = f"Outer(inner=TestAsDictWithoutDefaultsAndReprWithoutDefaults.test_nested_without_recursive.<locals>.Inner(x={x}))"
        assert repr_res == repr_exp

    def test_ib_async(self) -> None:
        fut = Future(
            conId=495512557,
            symbol="ES",
            lastTradeDateOrContractMonth="20241220",
            strike=0.0,
            right="",
            multiplier="50",
            exchange="",
            primaryExchange="",
            currency="USD",
            localSymbol="ESZ4",
            tradingClass="ES",
            includeExpired=False,
            secIdType="",
            secId="",
            description="",
            issuerId="",
            comboLegsDescrip="",
            comboLegs=[],
            deltaNeutralContract=None,
        )
        result = asdict_without_defaults(fut)
        expected = {
            "secType": "FUT",
            "conId": 495512557,
            "symbol": "ES",
            "lastTradeDateOrContractMonth": "20241220",
            "multiplier": "50",
            "currency": "USD",
            "localSymbol": "ESZ4",
            "tradingClass": "ES",
        }
        assert result == expected


class TestMappingToDataclass:
    @given(value=integers())
    def test_case_sensitive(self, *, value: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        obj = mapping_to_dataclass(Example, {"x": value})
        expected = Example(x=value)
        assert obj == expected

    @given(key=sampled_from(["x", "X"]), value=integers())
    def test_case_insensitive(self, *, key: str, value: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        obj = mapping_to_dataclass(Example, {key: value}, case_sensitive=False)
        expected = Example(x=value)
        assert obj == expected

    @given(value=text_ascii())
    def test_post(self, *, value: str) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: str

        obj = mapping_to_dataclass(Example, {"x": value}, post=lambda _, x: x.upper())
        expected = Example(x=value.upper())
        assert obj == expected

    @given(value=integers())
    def test_error_case_sensitive_empty_error(self, *, value: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        with raises(
            _MappingToDataclassEmptyError, match=r"Mapping .* does not contain 'x'"
        ):
            _ = mapping_to_dataclass(Example, {"X": value})

    @given(value=integers())
    def test_error_case_insensitive_empty_error(self, *, value: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        with raises(
            _MappingToDataclassEmptyError,
            match=r"Mapping .* does not contain 'x' \(modulo case\)",
        ):
            _ = mapping_to_dataclass(Example, {"y": value}, case_sensitive=False)

    @given(value1=integers(), value2=integers())
    def test_error_case_insensitive_non_unique_error(
        self, *, value1: int, value2: int
    ) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        with raises(
            _MappingToDataclassCaseInsensitiveNonUniqueError,
            match=re.compile(
                r"Mapping .* must contain 'x' exactly once \(modulo case\); got 'x', 'X' and perhaps more",
                flags=DOTALL,
            ),
        ):
            _ = mapping_to_dataclass(
                Example, {"x": value1, "X": value2}, case_sensitive=False
            )


class TestReplaceNonSentinel:
    def test_main(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example()
        assert obj.x == 0
        obj1 = replace_non_sentinel(obj, x=1)
        assert obj1.x == 1
        obj2 = replace_non_sentinel(obj1, x=sentinel)
        assert obj2.x == 1

    def test_in_place(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example()
        assert obj.x == 0
        replace_non_sentinel(obj, x=1, in_place=True)
        assert obj.x == 1
        replace_non_sentinel(obj, x=sentinel, in_place=True)
        assert obj.x == 1


class TestReprWithoutDefaults:
    def test_overriding_repr(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

            @override
            def __repr__(self) -> str:
                return repr_without_defaults(self)

        obj = Example()
        result = repr(obj)
        expected = "Example()"
        assert result == expected

    @given(x=integers())
    def test_non_repr_field(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = field(default=0, repr=False)

        obj = Example(x=x)
        result = repr_without_defaults(obj)
        expected = "Example()"
        assert result == expected

    @given(x=integers(), y=integers())
    def test_ignore(self, *, x: int, y: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int
            y: int

        obj = Example(x=x, y=y)
        result = repr_without_defaults(obj, ignore="x")
        expected = f"Example(y={y})"
        assert result == expected


class TestYieldFields:
    def test_class_with_none_type_no_default(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: None

        result = one(yield_fields(Example))
        expected = _YieldFieldsClass(name="x", type_=NoneType, kw_only=True)
        assert result == expected

    def test_class_with_none_type_and_default(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: None = None

        result = one(yield_fields(Example))
        expected = _YieldFieldsClass(
            name="x", type_=NoneType, default=None, kw_only=True
        )
        assert result == expected

    def test_class_with_int_type(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        result = one(yield_fields(Example))
        expected = _YieldFieldsClass(name="x", type_=int, kw_only=True)
        assert result == expected

    def test_class_with_list_int_type(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: list[int] = field(default_factory=list)

        result = one(yield_fields(Example))
        expected = _YieldFieldsClass(
            name="x", type_=list[int], default_factory=list, kw_only=True
        )
        assert result == expected
        assert is_list_type(result.type_)
        assert get_args(result.type_) == (int,)

    def test_class_nested(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner

        result = one(yield_fields(Outer, localns=locals()))
        expected = _YieldFieldsClass(name="inner", type_=Inner, kw_only=True)
        assert result == expected
        assert result.type_ is Inner

    def test_class_literal(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: TruthLit

        result = one(yield_fields(Example, globalns=globals()))
        expected = _YieldFieldsClass(name="truth", type_=TruthLit, kw_only=True)

        assert result == expected
        assert is_literal_type(result.type_)
        assert get_args(result.type_) == ("true", "false")

    def test_class_literal_nullable(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            truth: TruthLit | None = None

        result = one(yield_fields(Example, globalns=globals()))
        expected = _YieldFieldsClass(
            name="truth", type_=TruthLit | None, default=None, kw_only=True
        )
        assert result == expected
        assert is_optional_type(result.type_)
        args = get_args(result.type_)
        assert args == (Literal["true", "false"],)
        arg = one(args)
        assert get_args(arg) == ("true", "false")

    def test_class_literal_defined_elsewhere(self) -> None:
        result = one(yield_fields(DataClass3))
        expected = _YieldFieldsClass(
            name="truth", type_=Literal["true", "false"], kw_only=True
        )
        assert result == expected

    def test_class_orjson_log_record(self) -> None:
        result = list(yield_fields(OrjsonLogRecord, globalns=globals()))
        exp_head = [
            _YieldFieldsClass(name="name", type_=str, kw_only=True),
            _YieldFieldsClass(name="message", type_=str, kw_only=True),
            _YieldFieldsClass(name="level", type_=int, kw_only=True),
        ]
        assert result[:3] == exp_head
        exp_tail = [
            _YieldFieldsClass(
                name="extra", type_=StrMapping | None, default=None, kw_only=True
            ),
            _YieldFieldsClass(
                name="log_file", type_=Path | None, default=None, kw_only=True
            ),
            _YieldFieldsClass(
                name="log_file_line_num", type_=int | None, default=None, kw_only=True
            ),
        ]
        assert result[-3:] == exp_tail

    def test_instance(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: None = None

        obj = Example()
        result = one(yield_fields(obj))
        expected = _YieldFieldsInstance(
            name="x", value=None, type_=NoneType, default=None, kw_only=True
        )
        assert result == expected

    @given(x=integers())
    def test_instance_is_default_value_with_no_default(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        obj = Example(x=x)
        field = one(yield_fields(obj))
        assert not field.equals_default()

    @given(x=integers())
    def test_instance_is_default_value_with_default(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example(x=x)
        field = one(yield_fields(obj))
        result = field.equals_default()
        expected = x == 0
        assert result is expected

    @given(x=lists(integers()))
    def test_instance_is_default_value_with_default_factory(
        self, *, x: list[int]
    ) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: list[int] = field(default_factory=list)

        obj = Example(x=x)
        fld = one(yield_fields(obj))
        result = fld.equals_default()
        expected = x == []
        assert result is expected

    def test_error(self) -> None:
        with raises(
            YieldFieldsError,
            match="Object must be a dataclass instance or class; got None",
        ):
            _ = list(yield_fields(cast(Any, None)))



from __future__ import annotations

from asyncio import iscoroutinefunction
from collections.abc import Callable
from functools import partial, wraps
from typing import Any, TypeVar, cast, overload

from beartype import beartype

from utilities.typing import contains_self, get_type_hints

_F = TypeVar("_F", bound=Callable[..., Any])


@overload
def beartype_cond(
    func: _F,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> _F: ...
@overload
def beartype_cond(
    func: None = None,
    /,
    *,
    setup: Callable[[], bool] | None = ...,
    runtime: Callable[[], bool] | None = ...,
) -> Callable[[_F], _F]: ...
def beartype_cond(
    func: _F | None = None,
    /,
    *,
    setup: Callable[[], bool] | None = None,
    runtime: Callable[[], bool] | None = None,
) -> _F | Callable[[_F], _F]:
    """Apply `beartype` conditionally."""
    if func is None:
        result = partial(beartype_cond, setup=setup, runtime=runtime)
        return cast(Callable[[_F], _F], result)

    if (setup is not None) and not setup():
        return func

    if any(map(contains_self, get_type_hints(func).values())):
        decorated = func
    else:
        decorated = beartype(func)

    if runtime is None:
        return decorated

    if not iscoroutinefunction(func):

        @wraps(func)
        def beartype_sync(*args: Any, **kwargs: Any) -> Any:
            if runtime():
                return decorated(*args, **kwargs)
            return func(*args, **kwargs)

        return cast(_F, beartype_sync)

    @wraps(func)
    async def beartype_async(*args: Any, **kwargs: Any) -> Any:
        if runtime():
            return await decorated(*args, **kwargs)
        return await func(*args, **kwargs)

    return cast(_F, beartype_async)


__all__ = ["beartype_cond"]



from __future__ import annotations

from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING, Literal, TypedDict
from warnings import catch_warnings, filterwarnings

if TYPE_CHECKING:
    from collections.abc import Iterator


@contextmanager
def catch_warnings_as_errors(
    *,
    message: str = "",
    category: type[Warning] | tuple[type[Warning], ...] | None = None,
) -> Iterator[None]:
    """Catch warnings as errors."""
    with _handle_warnings("error", message=message, category=category):
        yield


@contextmanager
def suppress_warnings(
    *,
    message: str = "",
    category: type[Warning] | tuple[type[Warning], ...] | None = None,
) -> Iterator[None]:
    """Suppress warnings."""
    with _handle_warnings("ignore", message=message, category=category):
        yield


_ActionKind = Literal["error", "ignore"]


def _handle_warnings(
    action: _ActionKind,
    /,
    *,
    message: str = "",
    category: type[Warning] | tuple[type[Warning], ...] | None = None,
) -> ExitStack:
    """Handle a set of warnings."""
    stack = ExitStack()
    categories = category if isinstance(category, tuple) else [category]
    for cat in categories:
        cm = _handle_warnings_1(action, message=message, category=cat)
        stack.enter_context(cm)
    return stack


@contextmanager
def _handle_warnings_1(
    action: _ActionKind, /, *, message: str = "", category: type[Warning] | None = None
) -> Iterator[None]:
    """Handle one set of warnings."""

    class Kwargs(TypedDict, total=False):
        category: type[Warning]

    with catch_warnings():
        kwargs: Kwargs = {} if category is None else {"category": category}
        filterwarnings(action, message=message, **kwargs)
        yield


__all__ = ["catch_warnings_as_errors", "suppress_warnings"]



from __future__ import annotations

import datetime as dt
import re
from operator import eq
from re import DOTALL
from typing import TYPE_CHECKING, Any

import polars as pl
import sqlalchemy
from hypothesis import Phase, given, settings
from hypothesis.strategies import (
    DataObject,
    SearchStrategy,
    booleans,
    data,
    dates,
    datetimes,
    floats,
    integers,
    just,
    lists,
    none,
    sampled_from,
    sets,
)
from polars import (
    Binary,
    DataFrame,
    Datetime,
    Decimal,
    Duration,
    Float64,
    Int32,
    Int64,
    UInt32,
    UInt64,
    col,
    when,
)
from polars.testing import assert_frame_equal
from pytest import mark, param, raises
from sqlalchemy import (
    BIGINT,
    BINARY,
    BOOLEAN,
    CHAR,
    CLOB,
    DATE,
    DATETIME,
    DECIMAL,
    DOUBLE,
    DOUBLE_PRECISION,
    FLOAT,
    INT,
    INTEGER,
    NCHAR,
    NUMERIC,
    NVARCHAR,
    REAL,
    SMALLINT,
    TEXT,
    TIME,
    TIMESTAMP,
    UUID,
    VARBINARY,
    VARCHAR,
    BigInteger,
    Column,
    DateTime,
    Double,
    Float,
    Integer,
    Interval,
    LargeBinary,
    MetaData,
    Numeric,
    Select,
    SmallInteger,
    Table,
    Text,
    Unicode,
    UnicodeText,
    Uuid,
    select,
)
from sqlalchemy.exc import DuplicateColumnError, OperationalError, ProgrammingError

from tests.conftest import FLAKY
from tests.test_sqlalchemy import _table_names, _upsert_lists
from utilities.datetime import is_equal_mod_tz
from utilities.hypothesis import sqlalchemy_engines, text_ascii
from utilities.math import is_equal
from utilities.polars import DatetimeUTC, check_polars_dataframe
from utilities.sqlalchemy import ensure_tables_created
from utilities.sqlalchemy_polars import (
    InsertDataFrameError,
    _insert_dataframe_check_df_and_db_types,
    _insert_dataframe_map_df_column_to_table_column_and_type,
    _insert_dataframe_map_df_column_to_table_schema,
    _insert_dataframe_map_df_schema_to_table,
    _InsertDataFrameMapDFColumnToTableColumnAndTypeError,
    _InsertDataFrameMapDFColumnToTableSchemaError,
    _select_to_dataframe_apply_snake,
    _select_to_dataframe_check_duplicates,
    _select_to_dataframe_map_select_to_df_schema,
    _select_to_dataframe_map_table_column_type_to_dtype,
    _select_to_dataframe_yield_selects_with_in_clauses,
    insert_dataframe,
    select_to_dataframe,
)
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence

    from polars._typing import PolarsDataType
    from polars.datatypes import DataTypeClass
    from sqlalchemy.ext.asyncio import AsyncEngine


_CASES_SELECT: list[
    tuple[SearchStrategy[Any], PolarsDataType, Any, Callable[[Any, Any], bool]]
] = [
    (booleans() | none(), pl.Boolean, sqlalchemy.Boolean, eq),
    (dates() | none(), pl.Date, sqlalchemy.Date, eq),
    (datetimes() | none(), Datetime, DateTime, eq),
    (
        datetimes(timezones=just(UTC)) | none(),
        DatetimeUTC,
        DateTime(timezone=True),
        is_equal_mod_tz,
    ),
    (floats(allow_nan=False) | none(), Float64, Float, is_equal),
    (integers(-10, 10) | none(), Int64, Integer, eq),
    (text_ascii() | none(), pl.String, sqlalchemy.String, eq),
]
_CASES_INSERT: list[
    tuple[SearchStrategy[Any], PolarsDataType, Any, Callable[[Any, Any], bool]]
] = [*_CASES_SELECT, (integers(-10, 10) | none(), Int32, Integer, eq)]


class TestInsertDataFrame:
    @FLAKY
    @given(data=data(), name=_table_names(), case=sampled_from(_CASES_INSERT))
    @settings(phases={Phase.generate})
    async def test_main(
        self,
        *,
        data: DataObject,
        name: str,
        case: tuple[
            SearchStrategy[Any], PolarsDataType, Any, Callable[[Any, Any], bool]
        ],
    ) -> None:
        strategy, pl_dtype, col_type, check = case
        values = data.draw(lists(strategy, max_size=100))
        df = DataFrame({"value": values}, schema={"value": pl_dtype})
        table = self._make_table(name, col_type)
        engine = await sqlalchemy_engines(data, table)
        await insert_dataframe(df, table, engine)
        sel = select(table.c["value"])
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).scalars().all()
        for r, v in zip(results, values, strict=True):
            assert ((r is None) == (v is None)) or check(r, v)

    @FLAKY
    @given(data=data(), name=_table_names())
    @settings(phases={Phase.generate})
    async def test_assume_exists_with_empty(
        self, *, data: DataObject, name: str
    ) -> None:
        df = DataFrame(schema={"value": pl.Boolean})
        table = self._make_table(name, sqlalchemy.Boolean)
        engine = await sqlalchemy_engines(data, table)
        await ensure_tables_created(engine, table)
        await insert_dataframe(df, table, engine, assume_tables_exist=True)

    @FLAKY
    @given(data=data(), name=_table_names(), value=booleans())
    @settings(phases={Phase.generate})
    async def test_assume_exists_with_non_empty(
        self, *, data: DataObject, name: str, value: bool
    ) -> None:
        df = DataFrame([(value,)], schema={"value": pl.Boolean})
        table = self._make_table(name, sqlalchemy.Boolean)
        engine = await sqlalchemy_engines(data, table)
        with raises(
            (OperationalError, ProgrammingError), match="(no such table|does not exist)"
        ):
            await insert_dataframe(df, table, engine, assume_tables_exist=True)

    @FLAKY
    @given(data=data(), name=_table_names())
    @settings(phases={Phase.generate})
    async def test_empty(self, *, data: DataObject, name: str) -> None:
        df = DataFrame(schema={"value": pl.Boolean})
        table = self._make_table(name, sqlalchemy.Boolean)
        engine = await sqlalchemy_engines(data, table)
        await insert_dataframe(df, table, engine, assume_tables_exist=False)
        sel = select(table.c["value"])
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).scalars().all()
        assert results == []

    @FLAKY
    @given(data=data(), name=_table_names())
    @settings(phases={Phase.generate})
    async def test_upsert(self, *, data: DataObject, name: str) -> None:
        values = data.draw(_upsert_lists())
        df = DataFrame(
            values,
            schema={"id_": Int64, "init": pl.Boolean, "post": pl.Boolean},
            orient="row",
        )
        table = self._make_table(name, sqlalchemy.Boolean)
        engine = await sqlalchemy_engines(data, table)
        await insert_dataframe(
            df.select("id_", col("init").alias("value")),
            table,
            engine,
            upsert="selected",
        )
        sel = select(table)
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).all()
        assert set(results) == set(df.select("id_", "init").rows())
        await insert_dataframe(
            df.select("id_", col("post").alias("value")).drop_nulls(),
            table,
            engine,
            upsert="selected",
        )
        async with engine.begin() as conn:
            results = (await conn.execute(sel)).all()
        expected = df.select(
            "id_", when(col("post").is_null()).then("init").otherwise("post")
        )
        assert set(results) == set(expected.rows())

    @FLAKY
    @given(data=data(), name=_table_names(), value=booleans())
    @settings(phases={Phase.generate})
    async def test_error(self, *, data: DataObject, name: str, value: bool) -> None:
        df = DataFrame([(value,)], schema={"other": pl.Boolean})
        table = self._make_table(name, sqlalchemy.Boolean)
        engine = await sqlalchemy_engines(data, table)
        with raises(
            InsertDataFrameError,
            match="Non-empty DataFrame must resolve to at least 1 item",
        ):
            _ = await insert_dataframe(df, table, engine)

    def _make_table(self, name: str, type_: Any, /) -> Table:
        return Table(
            name,
            MetaData(),
            Column("id_", Integer, primary_key=True),
            Column("value", type_),
        )


class TestInsertDataFrameCheckDFAndDBTypes:
    @mark.parametrize(
        ("dtype", "db_col_type", "expected"),
        [
            param(pl.Boolean, bool, True),
            param(pl.Date, dt.date, True),
            param(pl.Date, dt.datetime, False),
            param(Datetime, dt.date, False),
            param(Datetime, dt.datetime, True),
            param(Float64, float, True),
            param(Float64, int, False),
            param(Int32, int, True),
            param(Int32, float, False),
            param(Int64, int, True),
            param(Int64, float, False),
            param(UInt32, int, True),
            param(UInt32, float, False),
            param(UInt64, int, True),
            param(UInt64, float, False),
            param(pl.String, str, True),
        ],
    )
    def test_main(
        self, *, dtype: PolarsDataType, db_col_type: type, expected: bool
    ) -> None:
        result = _insert_dataframe_check_df_and_db_types(dtype, db_col_type)
        assert result is expected


class TestInsertDataFrameMapDFColumnToTableColumnAndType:
    def test_main(self) -> None:
        schema = {"a": int, "b": float, "c": str}
        result = _insert_dataframe_map_df_column_to_table_column_and_type("b", schema)
        expected = ("b", float)
        assert result == expected

    @mark.parametrize("sr_name", [param("b"), param("B")])
    def test_snake(self, *, sr_name: str) -> None:
        schema = {"A": int, "B": float, "C": str}
        result = _insert_dataframe_map_df_column_to_table_column_and_type(
            sr_name, schema, snake=True
        )
        expected = ("B", float)
        assert result == expected

    @mark.parametrize("snake", [param(True), param(False)])
    def test_error_empty(self, *, snake: bool) -> None:
        schema = {"a": int, "b": float, "c": str}
        with raises(
            _InsertDataFrameMapDFColumnToTableColumnAndTypeError,
            match=r"Unable to map DataFrame column 'value' into table schema \{.*\} with snake=[True|False]",
        ):
            _ = _insert_dataframe_map_df_column_to_table_column_and_type(
                "value", schema, snake=snake
            )

    def test_error_non_unique(self) -> None:
        schema = {"a": int, "b": float, "B": float, "c": str}
        with raises(
            _InsertDataFrameMapDFColumnToTableColumnAndTypeError,
            match=re.compile(
                r"Unable to map DataFrame column 'b' into table schema \{.*\} with snake=True",
                flags=DOTALL,
            ),
        ):
            _ = _insert_dataframe_map_df_column_to_table_column_and_type(
                "b", schema, snake=True
            )


class TestInsertDataFrameMapDFColumnToTableSchema:
    def test_main(self) -> None:
        table_schema = {"a": int, "b": float, "c": str}
        result = _insert_dataframe_map_df_column_to_table_schema(
            "b", Float64, table_schema
        )
        assert result == "b"

    def test_error(self) -> None:
        table_schema = {"a": int, "b": float, "c": str}
        with raises(_InsertDataFrameMapDFColumnToTableSchemaError):
            _ = _insert_dataframe_map_df_column_to_table_schema(
                "b", Int64, table_schema
            )


class TestInsertDataFrameMapDFSchemaToTable:
    def test_default(self) -> None:
        df_schema = {"a": Int64, "b": Float64}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
        )
        result = _insert_dataframe_map_df_schema_to_table(df_schema, table)
        expected = {"a": "a", "b": "b"}
        assert result == expected

    def test_snake(self) -> None:
        df_schema = {"a": Int64, "b": Float64}
        table = Table(
            "example",
            MetaData(),
            Column("Id", Integer, primary_key=True),
            Column("A", Integer),
            Column("B", Float),
        )
        result = _insert_dataframe_map_df_schema_to_table(df_schema, table, snake=True)
        expected = {"a": "A", "b": "B"}
        assert result == expected

    def test_df_schema_has_extra_columns(self) -> None:
        df_schema = {"a": Int64, "b": Float64, "c": pl.String}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
        )
        result = _insert_dataframe_map_df_schema_to_table(df_schema, table)
        expected = {"a": "a", "b": "b"}
        assert result == expected

    def test_table_has_extra_columns(self) -> None:
        df_schema = {"a": Int64, "b": Float64}
        table = Table(
            "example",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("a", Integer),
            Column("b", Float),
            Column("c", sqlalchemy.String),
        )
        result = _insert_dataframe_map_df_schema_to_table(df_schema, table)
        expected = {"a": "a", "b": "b"}
        assert result == expected


class TestSelectToDataFrame:
    @FLAKY
    @given(
        data=data(),
        name=_table_names(),
        case=sampled_from([
            (strategy, pl_dtype, col_type)
            for (strategy, pl_dtype, col_type, _) in _CASES_SELECT
        ]),
    )
    @settings(phases={Phase.generate})
    async def test_main(
        self,
        *,
        data: DataObject,
        name: str,
        case: tuple[SearchStrategy[Any], PolarsDataType, Any],
    ) -> None:
        strategy, pl_dtype, col_type = case
        values = data.draw(lists(strategy, max_size=100))
        df = DataFrame({"value": values}, schema={"value": pl_dtype})
        table = self._make_table(name, col_type)
        engine = await sqlalchemy_engines(data, table)
        await insert_dataframe(df, table, engine)
        sel = select(table.c["value"])
        result = await select_to_dataframe(sel, engine)
        assert_frame_equal(result, df)

    @FLAKY
    @given(
        data=data(),
        name=_table_names(),
        values=lists(booleans() | none(), min_size=1, max_size=100),
        sr_name=sampled_from(["Value", "value"]),
    )
    @settings(phases={Phase.generate})
    async def test_snake(
        self, *, data: DataObject, name: str, values: list[bool], sr_name: str
    ) -> None:
        df = DataFrame({sr_name: values}, schema={sr_name: pl.Boolean})
        table = self._make_table(name, sqlalchemy.Boolean, title=True)
        engine = await sqlalchemy_engines(data, table)
        await insert_dataframe(df, table, engine, snake=True)
        sel = select(table.c["Value"])
        result = await select_to_dataframe(sel, engine, snake=True)
        expected = DataFrame({"value": values}, schema={"value": pl.Boolean})
        assert_frame_equal(result, expected)

    @FLAKY
    @given(
        data=data(),
        name=_table_names(),
        values=lists(integers(0, 100), min_size=1, max_size=100, unique=True),
        batch_size=integers(1, 10),
    )
    @settings(phases={Phase.generate})
    async def test_batch(
        self, *, data: DataObject, name: str, values: list[int], batch_size: int
    ) -> None:
        df, table, engine, sel = await self._prepare_feature_test(data, name, values)
        await insert_dataframe(df, table, engine)
        dfs = await select_to_dataframe(sel, engine, batch_size=batch_size)
        self._assert_batch_results(dfs, batch_size, values)

    @FLAKY
    @given(
        data=data(),
        name=_table_names(),
        values=lists(integers(0, 100), min_size=1, max_size=100, unique=True),
        in_clauses_chunk_size=integers(1, 10),
    )
    @settings(phases={Phase.generate})
    async def test_in_clauses_non_empty(
        self,
        *,
        data: DataObject,
        name: str,
        values: list[int],
        in_clauses_chunk_size: int,
    ) -> None:
        df, table, engine, sel = await self._prepare_feature_test(data, name, values)
        await insert_dataframe(df, table, engine)
        in_values = data.draw(sets(sampled_from(values)))
        df = await select_to_dataframe(
            sel,
            engine,
            in_clauses=(table.c["value"], in_values),
            in_clauses_chunk_size=in_clauses_chunk_size,
        )
        check_polars_dataframe(df, height=len(in_values), schema_list={"value": Int64})
        assert set(df["value"].to_list()) == in_values

    @FLAKY
    @given(data=data(), name=_table_names())
    @settings(phases={Phase.generate})
    async def test_in_clauses_empty(self, *, data: DataObject, name: str) -> None:
        table = self._make_table(name, Integer)
        engine = await sqlalchemy_engines(data, table)
        await ensure_tables_created(engine, table)
        sel = select(table.c["value"])
        df = await select_to_dataframe(sel, engine, in_clauses=(table.c["value"], []))
        check_polars_dataframe(df, height=0, schema_list={"value": Int64})

    @FLAKY
    @given(
        data=data(),
        name=_table_names(),
        values=lists(integers(0, 100), min_size=1, max_size=100, unique=True),
        batch_size=integers(1, 10),
        in_clauses_chunk_size=integers(1, 10),
    )
    @settings(phases={Phase.generate})
    async def test_batch_and_in_clauses(
        self,
        *,
        data: DataObject,
        name: str,
        values: list[int],
        batch_size: int,
        in_clauses_chunk_size: int,
    ) -> None:
        df, table, engine, sel = await self._prepare_feature_test(data, name, values)
        await insert_dataframe(df, table, engine)
        in_values = data.draw(sets(sampled_from(values)))
        async_dfs = await select_to_dataframe(
            sel,
            engine,
            batch_size=batch_size,
            in_clauses=(table.c["value"], in_values),
            in_clauses_chunk_size=in_clauses_chunk_size,
        )
        sync_dfs = [df async for df in async_dfs]
        max_height = batch_size * in_clauses_chunk_size
        self._assert_batch_results(sync_dfs, max_height, in_values)

    def _assert_batch_results(
        self, dfs: Iterable[DataFrame], max_height: int, values: Iterable[int], /
    ) -> None:
        seen: set[int] = set()
        values = set(values)
        for df_i in dfs:
            check_polars_dataframe(
                df_i, max_height=max_height, schema_list={"value": Int64}
            )
            assert df_i["value"].is_in(values).all()
            seen.update(df_i["value"].to_list())
        assert seen == values

    def _make_table(self, name: str, type_: Any, /, *, title: bool = False) -> Table:
        return Table(
            name,
            MetaData(),
            Column("Id" if title else "id", Integer, primary_key=True),
            Column("Value" if title else "value", type_),
        )

    async def _prepare_feature_test(
        self, data: DataObject, name: str, values: Sequence[int], /
    ) -> tuple[DataFrame, Table, AsyncEngine, Select[Any]]:
        df = DataFrame({"value": values}, schema={"value": Int64})
        table = self._make_table(name, Integer)
        engine = await sqlalchemy_engines(data, table)
        sel = select(table.c["value"])
        return df, table, engine, sel


class TestSelectToDataFrameApplySnake:
    def test_main(self) -> None:
        table = Table(
            "example",
            MetaData(),
            Column("Id", Integer, primary_key=True),
            Column("Value", sqlalchemy.Boolean),
        )
        sel = select(table)
        res = _select_to_dataframe_apply_snake(sel)
        expected = ["id", "value"]
        for column, exp in zip(res.selected_columns, expected, strict=True):
            assert column.name == exp


class TestSelectToDataFrameCheckDuplicates:
    def test_error(self) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        sel = select(table.c["id"], table.c["id"])
        with raises(
            DuplicateColumnError,
            match=r"Columns must not contain duplicates; got \{'id': 2\}",
        ):
            _select_to_dataframe_check_duplicates(sel.selected_columns)


class TestSelectToDataFrameMapSelectToDFSchema:
    def test_main(self) -> None:
        table = Table("example", MetaData(), Column("id", Integer, primary_key=True))
        sel = select(table.c["id"])
        schema = _select_to_dataframe_map_select_to_df_schema(sel)
        expected = {"id": Int64}
        assert schema == expected


class TestSelectToDataFrameMapTableColumnTypeToDType:
    @mark.parametrize(
        ("col_type", "expected"),
        [
            param(BigInteger, Int64),
            param(BIGINT, Int64),
            param(BINARY, Binary),
            param(sqlalchemy.Boolean, pl.Boolean),
            param(BOOLEAN, pl.Boolean),
            param(CHAR, pl.String),
            param(CLOB, pl.String),
            param(sqlalchemy.Date, pl.Date),
            param(DATE, pl.Date),
            param(DECIMAL, Decimal),
            param(Double, Float64),
            param(DOUBLE, Float64),
            param(DOUBLE_PRECISION, Float64),
            param(Float, Float64),
            param(FLOAT, Float64),
            param(INT, Int64),
            param(Integer, Int64),
            param(INTEGER, Int64),
            param(Interval, Duration),
            param(LargeBinary, Binary),
            param(NCHAR, pl.String),
            param(Numeric, Decimal),
            param(NUMERIC, Decimal),
            param(NVARCHAR, pl.String),
            param(REAL, Float64),
            param(SMALLINT, Int64),
            param(SmallInteger, Int64),
            param(sqlalchemy.String, pl.String),
            param(TEXT, pl.String),
            param(Text, pl.String),
            param(TIME, pl.Time),
            param(sqlalchemy.Time, pl.Time),
            param(Unicode, pl.String),
            param(UnicodeText, pl.String),
            param(Uuid, pl.String),
            param(UUID, pl.String),
            param(VARBINARY, Binary),
            param(VARCHAR, pl.String),
        ],
    )
    @mark.parametrize("use_inst", [param(True), param(False)])
    def test_main(
        self, *, col_type: Any, use_inst: bool, expected: DataTypeClass
    ) -> None:
        col_type_use = col_type() if use_inst else col_type
        dtype = _select_to_dataframe_map_table_column_type_to_dtype(col_type_use)
        assert isinstance(dtype, type)
        assert issubclass(dtype, expected)

    @mark.parametrize("col_type", [param(DATETIME), param(DateTime), param(TIMESTAMP)])
    @mark.parametrize("timezone", [param(None), param(True), param(False)])
    def test_datetime(self, *, col_type: Any, timezone: bool | None) -> None:
        col_type_use = col_type if timezone is None else col_type(timezone=timezone)
        dtype = _select_to_dataframe_map_table_column_type_to_dtype(col_type_use)
        assert isinstance(dtype, Datetime)


class TestSelectToDataFrameYieldSelectsWithInClauses:
    @given(
        data=data(),
        name=_table_names(),
        values=sets(integers(), max_size=100),
        in_clauses_chunk_size=integers(1, 10) | none(),
        chunk_size_frac=floats(0.1, 10.0),
    )
    async def test_main(
        self,
        *,
        data: DataObject,
        name: str,
        values: set[int],
        in_clauses_chunk_size: int | None,
        chunk_size_frac: float,
    ) -> None:
        table = Table(name, MetaData(), Column("id", Integer, primary_key=True))
        engine = await sqlalchemy_engines(data, table)
        sel = select(table.c["id"])
        sels = _select_to_dataframe_yield_selects_with_in_clauses(
            sel,
            engine,
            (table.c["id"], values),
            in_clauses_chunk_size=in_clauses_chunk_size,
            chunk_size_frac=chunk_size_frac,
        )
        for sel in sels:
            assert isinstance(sel, Select)



from __future__ import annotations

import datetime as dt
import re
from collections.abc import Callable, Mapping
from contextlib import suppress
from dataclasses import dataclass, field
from enum import Enum, unique
from functools import partial, reduce
from itertools import chain
from logging import Formatter, LogRecord
from math import isinf, isnan
from operator import or_
from pathlib import Path
from re import Pattern
from typing import TYPE_CHECKING, Any, Literal, TypeAlias, assert_never
from uuid import UUID

from orjson import (
    OPT_PASSTHROUGH_DATACLASS,
    OPT_PASSTHROUGH_DATETIME,
    OPT_SORT_KEYS,
    dumps,
    loads,
)
from typing_extensions import override

from utilities.concurrent import concurrent_map
from utilities.dataclasses import asdict_without_defaults
from utilities.functions import ensure_class
from utilities.iterables import OneEmptyError, one
from utilities.math import MAX_INT64, MIN_INT64
from utilities.types import Dataclass, PathLike, StrMapping
from utilities.uuid import UUID_PATTERN
from utilities.version import Version, parse_version
from utilities.whenever import (
    parse_date,
    parse_local_datetime,
    parse_time,
    parse_timedelta,
    parse_zoned_datetime,
    serialize_date,
    serialize_local_datetime,
    serialize_time,
    serialize_timedelta,
    serialize_zoned_datetime,
)

if TYPE_CHECKING:
    from collections.abc import Set as AbstractSet
    from logging import _FormatStyle

    from utilities.concurrent import Parallelism


# serialize


@unique
class _Prefixes(Enum):
    dataclass = "dc"
    date = "d"
    datetime = "dt"
    enum = "e"
    float_ = "fl"
    frozenset_ = "fr"
    list_ = "l"
    nan = "nan"
    path = "p"
    pos_inf = "pos_inf"
    neg_inf = "neg_inf"
    set_ = "s"
    timedelta = "td"
    time = "tm"
    tuple_ = "tu"
    unserializable = "un"
    uuid = "uu"
    version = "v"


_DataclassFinalHook: TypeAlias = Callable[[type[Dataclass], StrMapping], StrMapping]
_ErrorMode: TypeAlias = Literal["raise", "drop", "str"]


@dataclass(kw_only=True, slots=True)
class Unserializable:
    """An unserialiable object."""

    qualname: str
    repr: str
    str: str


def serialize(
    obj: Any,
    /,
    *,
    before: Callable[[Any], Any] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    dataclass_final_hook: _DataclassFinalHook | None = None,
) -> bytes:
    """Serialize an object."""
    obj_use = _pre_process(
        obj,
        before=before,
        globalns=globalns,
        localns=localns,
        dataclass_final_hook=dataclass_final_hook,
    )
    return dumps(
        obj_use,
        option=OPT_PASSTHROUGH_DATACLASS | OPT_PASSTHROUGH_DATETIME | OPT_SORT_KEYS,
    )


def _pre_process(
    obj: Any,
    /,
    *,
    before: Callable[[Any], Any] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    dataclass_final_hook: _DataclassFinalHook | None = None,
    error: _ErrorMode = "raise",
) -> Any:
    if before is not None:
        obj = before(obj)
    pre = partial(
        _pre_process,
        before=before,
        globalns=globalns,
        localns=localns,
        dataclass_final_hook=dataclass_final_hook,
        error=error,
    )
    match obj:
        # singletons
        case dt.datetime():
            if obj.tzinfo is None:
                ser = serialize_local_datetime(obj)
            elif obj.tzinfo is dt.UTC:
                ser = serialize_zoned_datetime(obj).replace("UTC", "dt.UTC")
            else:
                ser = serialize_zoned_datetime(obj)
            return f"[{_Prefixes.datetime.value}]{ser}"
        case dt.date():  # after datetime
            ser = serialize_date(obj)
            return f"[{_Prefixes.date.value}]{ser}"
        case dt.time():
            ser = serialize_time(obj)
            return f"[{_Prefixes.time.value}]{ser}"
        case dt.timedelta():
            ser = serialize_timedelta(obj)
            return f"[{_Prefixes.timedelta.value}]{ser}"
        case float():
            if isinf(obj) or isnan(obj):
                return f"[{_Prefixes.float_.value}]{obj}"
            return obj
        case int():
            if MIN_INT64 <= obj <= MAX_INT64:
                return obj
            raise _SerializeIntegerError(obj=obj)
        case UUID():
            return f"[{_Prefixes.uuid.value}]{obj}"
        case Path():
            ser = str(obj)
            return f"[{_Prefixes.path.value}]{ser}"
        case str():
            return obj
        case Version():
            ser = str(obj)
            return f"[{_Prefixes.version.value}]{ser}"
        # contains
        case Dataclass():
            obj_as_dict = asdict_without_defaults(
                obj,
                globalns=globalns,
                localns=localns,
                final=partial(_dataclass_final, hook=dataclass_final_hook),
            )
            return pre(obj_as_dict)
        case dict():
            return {k: pre(v) for k, v in obj.items()}
        case Enum():
            return {
                f"[{_Prefixes.enum.value}|{type(obj).__qualname__}]": pre(obj.value)
            }
        case frozenset():
            return _pre_process_container(
                obj,
                frozenset,
                _Prefixes.frozenset_,
                before=before,
                globalns=globalns,
                localns=localns,
                dataclass_final_hook=dataclass_final_hook,
            )
        case list():
            return _pre_process_container(
                obj,
                list,
                _Prefixes.list_,
                before=before,
                globalns=globalns,
                localns=localns,
                dataclass_final_hook=dataclass_final_hook,
            )
        case set():
            return _pre_process_container(
                obj,
                set,
                _Prefixes.set_,
                before=before,
                globalns=globalns,
                localns=localns,
                dataclass_final_hook=dataclass_final_hook,
            )
        case tuple():
            return _pre_process_container(
                obj,
                tuple,
                _Prefixes.tuple_,
                before=before,
                globalns=globalns,
                localns=localns,
                dataclass_final_hook=dataclass_final_hook,
            )
        # other
        case _:
            unserializable = Unserializable(
                qualname=type(obj).__qualname__, repr=repr(obj), str=str(obj)
            )
            return pre(unserializable)


def _pre_process_container(
    obj: Any,
    cls: type[frozenset | list | set | tuple],
    prefix: _Prefixes,
    /,
    *,
    before: Callable[[Any], Any] | None = None,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    dataclass_final_hook: _DataclassFinalHook | None = None,
) -> Any:
    values = [
        _pre_process(
            o,
            before=before,
            globalns=globalns,
            localns=localns,
            dataclass_final_hook=dataclass_final_hook,
        )
        for o in obj
    ]
    if issubclass(cls, list) and issubclass(list, type(obj)):
        return values
    if issubclass(cls, type(obj)):
        key = f"[{prefix.value}]"
    else:
        key = f"[{prefix.value}|{type(obj).__qualname__}]"
    return {key: values}


def _dataclass_final(
    cls: type[Dataclass],
    mapping: StrMapping,
    /,
    *,
    hook: Callable[[type[Dataclass], StrMapping], StrMapping] | None = None,
) -> StrMapping:
    if hook is not None:
        mapping = hook(cls, mapping)
    return {f"[{_Prefixes.dataclass.value}|{cls.__qualname__}]": mapping}


@dataclass(kw_only=True, slots=True)
class SerializeError(Exception):
    obj: Any


@dataclass(kw_only=True, slots=True)
class _SerializeIntegerError(SerializeError):
    @override
    def __str__(self) -> str:
        return f"Integer {self.obj} is out of range"


# deserialize


def deserialize(
    data: bytes,
    /,
    *,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> Any:
    """Deserialize an object."""
    return _object_hook(loads(data), data=data, objects=objects, redirects=redirects)


_LOCAL_DATETIME_PATTERN = re.compile(
    r"^\["
    + _Prefixes.datetime.value
    + r"\](\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?)$"
)
_UUID_PATTERN = re.compile(r"^\[" + _Prefixes.uuid.value + r"\](" + UUID_PATTERN + ")$")
_ZONED_DATETIME_PATTERN = re.compile(
    r"^\["
    + _Prefixes.datetime.value
    + r"\](\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?[\+\-]\d{2}:\d{2}(?::\d{2})?\[(?!(?:dt\.)).+?\])$"
)
_ZONED_DATETIME_ALTERNATIVE_PATTERN = re.compile(
    r"^\["
    + _Prefixes.datetime.value
    + r"\](\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(?:\.\d{1,6})?\+00:00\[dt\.UTC\])$"
)


def _make_unit_pattern(prefix: _Prefixes, /) -> Pattern[str]:
    return re.compile(r"^\[" + prefix.value + r"\](.+)$")


(
    _DATE_PATTERN,
    _FLOAT_PATTERN,
    _PATH_PATTERN,
    _TIME_PATTERN,
    _TIMEDELTA_PATTERN,
    _VERSION_PATTERN,
) = map(
    _make_unit_pattern,
    [
        _Prefixes.date,
        _Prefixes.float_,
        _Prefixes.path,
        _Prefixes.time,
        _Prefixes.timedelta,
        _Prefixes.version,
    ],
)


def _make_container_pattern(prefix: _Prefixes, /) -> Pattern[str]:
    return re.compile(r"^\[" + prefix.value + r"(?:\|(.+))?\]$")


(
    _DATACLASS_PATTERN,
    _ENUM_PATTERN,
    _FROZENSET_PATTERN,
    _LIST_PATTERN,
    _SET_PATTERN,
    _TUPLE_PATTERN,
) = map(
    _make_container_pattern,
    [
        _Prefixes.dataclass,
        _Prefixes.enum,
        _Prefixes.frozenset_,
        _Prefixes.list_,
        _Prefixes.set_,
        _Prefixes.tuple_,
    ],
)


def _object_hook(
    obj: bool | float | str | dict[str, Any] | list[Any] | Dataclass | None,  # noqa: FBT001
    /,
    *,
    data: bytes,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> Any:
    match obj:
        case bool() | int() | float() | Dataclass() | None:
            return obj
        case str():
            if match := _DATE_PATTERN.search(obj):
                return parse_date(match.group(1))
            if match := _FLOAT_PATTERN.search(obj):
                return float(match.group(1))
            if match := _LOCAL_DATETIME_PATTERN.search(obj):
                return parse_local_datetime(match.group(1))
            if match := _PATH_PATTERN.search(obj):
                return Path(match.group(1))
            if match := _TIME_PATTERN.search(obj):
                return parse_time(match.group(1))
            if match := _TIMEDELTA_PATTERN.search(obj):
                return parse_timedelta(match.group(1))
            if match := _UUID_PATTERN.search(obj):
                return UUID(match.group(1))
            if match := _VERSION_PATTERN.search(obj):
                return parse_version(match.group(1))
            if match := _ZONED_DATETIME_PATTERN.search(obj):
                return parse_zoned_datetime(match.group(1))
            if match := _ZONED_DATETIME_ALTERNATIVE_PATTERN.search(obj):
                return parse_zoned_datetime(
                    match.group(1).replace("dt.UTC", "UTC")
                ).replace(tzinfo=dt.UTC)
            return obj
        case list():
            return [
                _object_hook(o, data=data, objects=objects, redirects=redirects)
                for o in obj
            ]
        case dict():
            if len(obj) == 1:
                key, value = one(obj.items())
                for cls, pattern in [
                    (frozenset, _FROZENSET_PATTERN),
                    (list, _LIST_PATTERN),
                    (set, _SET_PATTERN),
                    (tuple, _TUPLE_PATTERN),
                ]:
                    result = _object_hook_container(
                        key,
                        value,
                        cls,
                        pattern,
                        data=data,
                        objects=objects,
                        redirects=redirects,
                    )
                    if result is not None:
                        return result
                result = _object_hook_dataclass(
                    key, value, data=data, objects=objects, redirects=redirects
                )
                if result is not None:
                    return result
                result = _object_hook_enum(
                    key, value, data=data, objects=objects, redirects=redirects
                )
                if result is not None:
                    return result
                return {
                    k: _object_hook(v, data=data, objects=objects, redirects=redirects)
                    for k, v in obj.items()
                }
            return {
                k: _object_hook(v, data=data, objects=objects, redirects=redirects)
                for k, v in obj.items()
            }
        case _ as never:
            assert_never(never)


def _object_hook_container(
    key: str,
    value: Any,
    cls: type[Any],
    pattern: Pattern[str],
    /,
    *,
    data: bytes,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> Any:
    if not (match := pattern.search(key)):
        return None
    if (qualname := match.group(1)) is None:
        cls_use = cls
    else:
        cls_use = _object_hook_get_object(
            qualname, data=data, objects=objects, redirects=redirects
        )
    return cls_use(
        _object_hook(v, data=data, objects=objects, redirects=redirects) for v in value
    )


def _object_hook_get_object(
    qualname: str,
    /,
    *,
    data: bytes = b"",
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> type[Any]:
    if qualname == Unserializable.__qualname__:
        return Unserializable
    if (objects is None) and (redirects is None):
        raise _DeserializeNoObjectsError(data=data, qualname=qualname)
    if objects is not None:
        with suppress(OneEmptyError):
            return one(o for o in objects if o.__qualname__ == qualname)
    if redirects:
        with suppress(KeyError):
            return redirects[qualname]
    raise _DeserializeObjectNotFoundError(data=data, qualname=qualname)


def _object_hook_dataclass(
    key: str,
    value: Any,
    /,
    *,
    data: bytes,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> Any:
    if not (match := _DATACLASS_PATTERN.search(key)):
        return None
    cls = _object_hook_get_object(
        match.group(1), data=data, objects=objects, redirects=redirects
    )
    items = {
        k: _object_hook(v, data=data, objects=objects, redirects=redirects)
        for k, v in value.items()
    }
    return cls(**items)


def _object_hook_enum(
    key: str,
    value: Any,
    /,
    *,
    data: bytes,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> Any:
    if not (match := _ENUM_PATTERN.search(key)):
        return None
    cls: type[Enum] = _object_hook_get_object(
        match.group(1), data=data, objects=objects, redirects=redirects
    )
    value_use = _object_hook(value, data=data, objects=objects, redirects=redirects)
    return one(i for i in cls if i.value == value_use)


@dataclass(kw_only=True, slots=True)
class DeserializeError(Exception):
    data: bytes
    qualname: str


@dataclass(kw_only=True, slots=True)
class _DeserializeNoObjectsError(DeserializeError):
    @override
    def __str__(self) -> str:
        return f"Objects required to deserialize {self.qualname!r} from {self.data!r}"


@dataclass(kw_only=True, slots=True)
class _DeserializeObjectNotFoundError(DeserializeError):
    @override
    def __str__(self) -> str:
        return (
            f"Unable to find object to deserialize {self.qualname!r} from {self.data!r}"
        )


# logging

_LOG_RECORD_DEFAULT_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "taskName",
    "thread",
    "threadName",
}


class OrjsonFormatter(Formatter):
    """Formatter for JSON logs."""

    @override
    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: _FormatStyle = "%",
        validate: bool = True,
        /,
        *,
        defaults: StrMapping | None = None,
        before: Callable[[Any], Any] | None = None,
        globalns: StrMapping | None = None,
        localns: StrMapping | None = None,
        dataclass_final_hook: _DataclassFinalHook | None = None,
    ) -> None:
        super().__init__(fmt, datefmt, style, validate, defaults=defaults)
        self._before = before
        self._globalns = globalns
        self._localns = localns
        self._dataclass_final_hook = dataclass_final_hook

    @override
    def format(self, record: LogRecord) -> str:
        from tzlocal import get_localzone

        extra = {
            k: v
            for k, v in record.__dict__.items()
            if (k not in _LOG_RECORD_DEFAULT_ATTRS) and (not k.startswith("_"))
        }
        log_record = OrjsonLogRecord(
            name=record.name,
            level=record.levelno,
            path_name=Path(record.pathname),
            line_num=record.lineno,
            message=record.getMessage(),
            datetime=dt.datetime.fromtimestamp(record.created, tz=get_localzone()),
            func_name=record.funcName,
            extra=extra if len(extra) >= 1 else None,
        )
        return serialize(
            log_record,
            before=self._before,
            globalns=self._globalns,
            localns=self._localns,
            dataclass_final_hook=self._dataclass_final_hook,
        ).decode()


@dataclass(kw_only=True, slots=True)
class OrjsonLogRecord:
    """The log record as a dataclass."""

    name: str
    message: str
    level: int
    path_name: Path
    line_num: int
    datetime: dt.datetime
    func_name: str | None = None
    stack_info: str | None = None
    extra: StrMapping | None = None
    log_file: Path | None = None
    log_file_line_num: int | None = None


@dataclass(kw_only=True, slots=True)
class GetLogRecordsOutput:
    """A collection of outputs."""

    path: Path
    files: list[Path] = field(default_factory=list)
    num_files: int = 0
    num_files_ok: int = 0
    num_files_error: int = 0
    num_lines: int = 0
    num_lines_ok: int = 0
    num_lines_error: int = 0
    records: list[OrjsonLogRecord] = field(default_factory=list, repr=False)
    missing: set[str] = field(default_factory=set)
    other_errors: list[Exception] = field(default_factory=list)

    @property
    def frac_files_ok(self) -> float:
        return self.num_files_ok / self.num_files

    @property
    def frac_files_error(self) -> float:
        return self.num_files_error / self.num_files

    @property
    def frac_lines_ok(self) -> float:
        return self.num_lines_ok / self.num_lines

    @property
    def frac_lines_error(self) -> float:
        return self.num_lines_error / self.num_lines


def get_log_records(
    path: PathLike,
    /,
    *,
    parallelism: Parallelism = "processes",
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> GetLogRecordsOutput:
    """Get the log records under a directory."""
    path = Path(path)
    files = [p for p in path.iterdir() if p.is_file()]
    func = partial(_get_log_records_one, objects=objects, redirects=redirects)
    try:
        from utilities.pqdm import pqdm_map
    except ModuleNotFoundError:  # pragma: no cover
        outputs = concurrent_map(func, files, parallelism=parallelism)
    else:
        outputs = pqdm_map(func, files, parallelism=parallelism)
    return GetLogRecordsOutput(
        path=path,
        files=files,
        num_files=sum(o.num_files for o in outputs),
        num_files_ok=sum(o.num_files_ok for o in outputs),
        num_files_error=sum(o.num_files_error for o in outputs),
        num_lines=sum(o.num_lines for o in outputs),
        num_lines_ok=sum(o.num_lines_ok for o in outputs),
        num_lines_error=sum(o.num_lines_error for o in outputs),
        records=sorted(
            chain.from_iterable(o.records for o in outputs), key=lambda r: r.datetime
        ),
        missing=set(reduce(or_, (o.missing for o in outputs), set())),
        other_errors=list(chain.from_iterable(o.other_errors for o in outputs)),
    )


@dataclass(kw_only=True, slots=True)
class _GetLogRecordsOneOutput:
    path: Path
    num_files: int = 0
    num_files_ok: int = 0
    num_files_error: int = 0
    num_lines: int = 0
    num_lines_ok: int = 0
    num_lines_error: int = 0
    records: list[OrjsonLogRecord] = field(default_factory=list, repr=False)
    missing: set[str] = field(default_factory=set)
    other_errors: list[Exception] = field(default_factory=list, repr=False)


def _get_log_records_one(
    path: Path,
    /,
    *,
    objects: AbstractSet[type[Any]] | None = None,
    redirects: Mapping[str, type[Any]] | None = None,
) -> _GetLogRecordsOneOutput:
    path = Path(path)
    try:
        with path.open() as fh:
            lines = fh.readlines()
    except UnicodeDecodeError as error:  # skipif-ci-and-windows
        return _GetLogRecordsOneOutput(
            path=path, num_files=1, num_files_error=1, other_errors=[error]
        )
    num_lines_error = 0
    missing: set[str] = set()
    records: list[OrjsonLogRecord] = []
    errors: list[Exception] = []
    objects_use = {OrjsonLogRecord} | (set() if objects is None else objects)
    for i, line in enumerate(lines, start=1):
        try:
            result = deserialize(
                line.encode(), objects=objects_use, redirects=redirects
            )
            record = ensure_class(result, OrjsonLogRecord)
        except (_DeserializeNoObjectsError, _DeserializeObjectNotFoundError) as error:
            num_lines_error += 1
            missing.add(error.qualname)
        except Exception as error:  # noqa: BLE001
            num_lines_error += 1
            errors.append(error)
        else:
            record.log_file = path
            record.log_file_line_num = i
            records.append(record)
    return _GetLogRecordsOneOutput(
        path=path,
        num_files=1,
        num_files_ok=1,
        num_files_error=0,
        num_lines=len(lines),
        num_lines_ok=len(records),
        num_lines_error=num_lines_error,
        records=sorted(records, key=lambda r: r.datetime),
        missing=missing,
        other_errors=errors,
    )


__all__ = [
    "DeserializeError",
    "GetLogRecordsOutput",
    "OrjsonFormatter",
    "OrjsonLogRecord",
    "SerializeError",
    "deserialize",
    "get_log_records",
    "serialize",
]



from __future__ import annotations

UUID_PATTERN = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
UUID_EXACT_PATTERN = f"^{UUID_PATTERN}$"


__all__ = ["UUID_EXACT_PATTERN", "UUID_PATTERN"]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import partial
from os import environ
from re import IGNORECASE, search
from typing import TYPE_CHECKING, Any, TypeVar

from dotenv import dotenv_values
from typing_extensions import override

from utilities.dataclasses import (
    _MappingToDataclassEmptyError,
    _YieldFieldsClass,
    mapping_to_dataclass,
)
from utilities.enum import EnsureEnumError, ensure_enum
from utilities.functions import get_class_name
from utilities.git import get_repo_root
from utilities.iterables import MergeStrMappingsError, merge_str_mappings, one_str
from utilities.pathlib import PWD
from utilities.reprlib import get_repr
from utilities.types import Dataclass
from utilities.typing import get_args, is_literal_type

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from utilities.types import PathLike, StrMapping

_TDataclass = TypeVar("_TDataclass", bound=Dataclass)


def load_settings(
    cls: type[_TDataclass],
    /,
    *,
    cwd: PathLike = PWD,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
) -> _TDataclass:
    """Load a set of settings from the `.env` file."""
    path = get_repo_root(cwd=cwd).joinpath(".env")
    if not path.exists():
        raise _LoadSettingsFileNotFoundError(path=path) from None
    maybe_values_dotenv = dotenv_values(path)
    try:
        maybe_values = merge_str_mappings(
            maybe_values_dotenv, environ, case_sensitive=False
        )
    except MergeStrMappingsError as error:
        raise _LoadSettingsDuplicateKeysError(
            path=path, values=error.mapping, counts=error.counts
        ) from None
    values = {k: v for k, v in maybe_values.items() if v is not None}
    try:
        return mapping_to_dataclass(
            cls,
            values,
            globalns=globalns,
            localns=localns,
            case_sensitive=False,
            post=partial(_load_settings_post, path=path, values=values),
        )
    except _MappingToDataclassEmptyError as error:
        raise _LoadSettingsEmptyError(
            path=path, values=error.mapping, field=error.field
        ) from None


def _load_settings_post(
    field: _YieldFieldsClass[Any], value: Any, /, *, path: Path, values: StrMapping
) -> Any:
    type_ = field.type_
    if type_ is str:
        return value
    if type_ is bool:
        if value == "0" or search("false", value, flags=IGNORECASE):
            return False
        if value == "1" or search("true", value, flags=IGNORECASE):
            return True
        raise _LoadSettingsInvalidBoolError(
            path=path, values=values, field=field.name, value=value
        )
    if type_ is int:
        try:
            return int(value)
        except ValueError:
            raise _LoadSettingsInvalidIntError(
                path=path, values=values, field=field.name, value=value
            ) from None
    if isinstance(type_, type) and issubclass(type_, Enum):
        try:
            return ensure_enum(value, type_)
        except EnsureEnumError:
            raise _LoadSettingsInvalidEnumError(
                path=path, values=values, field=field.name, type_=type_, value=value
            ) from None
    if is_literal_type(type_):
        return one_str(get_args(type_), value, case_sensitive=False)
    raise _LoadSettingsTypeError(path=path, field=field.name, type=type_)


@dataclass(kw_only=True, slots=True)
class LoadSettingsError(Exception):
    path: Path


@dataclass(kw_only=True, slots=True)
class _LoadSettingsFileNotFoundError(LoadSettingsError):
    @override
    def __str__(self) -> str:
        return f"Path {str(self.path)!r} must exist"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsDuplicateKeysError(LoadSettingsError):
    values: StrMapping
    counts: Mapping[str, int]

    @override
    def __str__(self) -> str:
        return f"Mapping {get_repr(dict(self.values))} keys must not contain duplicates (modulo case); got {get_repr(self.counts)}"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsEmptyError(LoadSettingsError):
    values: StrMapping
    field: str

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} must exist (modulo case)"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsInvalidBoolError(LoadSettingsError):
    values: StrMapping
    field: str
    value: str

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} must contain a valid boolean; got {self.value!r}"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsInvalidIntError(LoadSettingsError):
    values: StrMapping
    field: str
    value: str

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} must contain a valid integer; got {self.value!r}"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsInvalidEnumError(LoadSettingsError):
    values: StrMapping
    field: str
    type_: type[Enum]
    value: str

    @override
    def __str__(self) -> str:
        type_ = get_class_name(self.type_)
        return f"Field {self.field!r} must contain a valid member of {type_!r}; got {self.value!r}"


@dataclass(kw_only=True, slots=True)
class _LoadSettingsTypeError(LoadSettingsError):
    field: str
    type: Any

    @override
    def __str__(self) -> str:
        return f"Field {self.field!r} has unsupported type {self.type!r}"


__all__ = ["LoadSettingsError", "load_settings"]



from __future__ import annotations

from hypothesis import example, given
from hypothesis.strategies import floats, integers, none

from utilities.cachetools import cache


class TestCache:
    @example(max_size=None, max_duration=None)
    @example(max_size=None, max_duration=1.0)
    @example(max_size=1, max_duration=None)
    @example(max_size=1, max_duration=1.0)
    @given(max_size=integers(1, 10) | none(), max_duration=floats(0.1, 10.0) | none())
    def test_main(self, *, max_size: int, max_duration: float) -> None:
        counter = 0

        @cache(max_size=max_size, max_duration=max_duration)
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        for _ in range(2):
            assert func(0) == 0
